/*
 *  libfb - FreeBASIC's runtime library
 *	Copyright (C) 2004-2011 The FreeBASIC development team.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  As a special exception, the copyright holders of this library give
 *  you permission to link this library with independent modules to
 *  produce an executable, regardless of the license terms of these
 *  independent modules, and to copy and distribute the resulting
 *  executable under terms of your choice, provided that you also meet,
 *  for each linked independent module, the terms and conditions of the
 *  license of that module. An independent module is a module which is
 *  not derived from or based on this library. If you modify this library,
 *  you may extend this exception to your version of the library, but
 *  you are not obligated to do so. If you do not wish to do so, delete
 *  this exception statement from your version.
 */

/*
 * time_time.c -- time$ function
 *
 * chng: nov/2004 written [v1ctor]
 *
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "fb.h"

/*:::::*/
FBCALL FBSTRING *fb_Time ( void )
{
	FBSTRING	*dst;
	time_t 		rawtime;
  	struct tm	*ptm;

	/* guard by global lock because time/localtime might not be thread-safe */
	FB_LOCK();

	rawtime = time( NULL );

	/* Note: localtime() may return NULL, as documented on MSDN and Linux man pages,
	   and it has been observed to do that on at least one FB user's Windows system,
	   because of a negative time_t value from time(). */
	if( ((ptm = localtime( &rawtime )) != NULL) &&
	    ((dst = fb_hStrAllocTemp( NULL, 2+1+2+1+2 )) != NULL) /* done last so it's not leaked */ ) {
		sprintf( dst->data, "%02d:%02d:%02d", ptm->tm_hour, ptm->tm_min, ptm->tm_sec );
	} else {
		dst = &__fb_ctx.null_desc;
	}

	FB_UNLOCK();

	return dst;
}

